@@ -630,9 +630,10 @@ int bch2_readdir(struct bch_fs *c, subvol_inum inum,
 
 			subvol_inum target;
 
-			bool need_second_pass = false;
+			bool need_second_pass = false, repaired_inode = false;
 			int ret2 = bch2_str_hash_check_key(trans, NULL, &bch2_dirent_hash_desc,
-							   hash_info, &iter, k, &need_second_pass) ?:
+							   hash_info, k,
+							   &need_second_pass, &repaired_inode) ?:
 				bch2_dirent_read_target(trans, inum, dirent, &target);
 			if (ret2 > 0)
 				continue;
@@ -675,7 +676,8 @@ int bch2_fsck_remove_dirent(struct btree_trans *trans, struct bpos pos)
 	struct bch_inode_unpacked dir_inode;
 	try(lookup_first_inode(trans, pos.inode, &dir_inode));
 
-	struct bch_hash_info dir_hash_info = bch2_hash_info_init(c, &dir_inode);
+	struct bch_hash_info dir_hash_info;
+	(bch2_hash_info_init(c, &dir_inode, &dir_hash_info));
 
 	CLASS(btree_iter, iter)(trans, BTREE_ID_dirents, pos, BTREE_ITER_intent);
 
