@@ -3,6 +3,7 @@
 #include <linux/log2.h>
 #include <linux/rcupdate.h>
 #include <linux/slab.h>
+#include <linux/version.h>
 #include <linux/vmalloc.h>
 #include "darray.h"
 
@@ -23,9 +24,15 @@ int __bch2_darray_resize_noprof(darray_char *d, size_t element_size, size_t new_
 			return -ENOMEM;
 
 		void *old = d->data;
+#if LINUX_VERSION_CODE < KERNEL_VERSION(6,18,0)
 		void *new = likely(bytes < INT_MAX)
 			? kvmalloc_noprof(bytes, gfp)
 			: vmalloc_noprof(bytes);
+#else
+		void *new = likely(bytes < INT_MAX)
+			? kvmalloc_node_align_noprof(bytes, 1, gfp, NUMA_NO_NODE)
+			: vmalloc_noprof(bytes);
+#endif
 		if (!new)
 			return -ENOMEM;
 
